/*
 * q_dsmark.c - Differentiated Services Marking qdisc
 *
 * Written 2001 by Werner Almesberger
 * Copyright 2001 EPFL-ICA, Network Robots
 */


#include <stddef.h>
#include <stdlib.h>
#include <assert.h>

#include "util.h"
#include "location.h"
#include "error.h"
#include "param.h"
#include "tree.h"
#include "op.h"
#include "tc.h"
#include "filter.h"
#include "qdisc_common.h"
#include "qdisc.h"
#include "ext.h"


#define __DEFAULT_PRM(f) f(mask)


/* ----- Class selection paths --------------------------------------------- */


/*
 * Known bugs and problems (of class selection paths):
 * - implementation is somewhat inefficient
 * - struct sel_ref and struct sel_path should be one structure
 * - does not detect cases where no default choice is needed (such is life ...)
 * - some error messages may appear a little obscure
 */

/*
 * Due to a bug in the dsmark API, we can only support 15 out of 16 possible
 * bits.
 */

#define DSMARK_INDEX_BITS 15


struct sel_path_elem {
    CLASS *class;
    struct sel_path_elem *next;
};

struct sel_path {
    struct sel_path_elem *path;
    int equals;			/* equal paths, >= 1 */
    int index;			/* nth such path, >= 0 */
    struct sel_path *next;
};

struct sel_ref {
    uint32_t *number;
    const DATA_LIST *list;
    struct sel_ref *next;
};

struct sel_class {
    CLASS *class;
    int number;
    struct sel_class *next;
};

struct sel_qdisc {
    QDISC *qdisc;
    struct sel_class *classes;
    int num_classes;
    int shift,bits;
    int curr_class;
    struct sel_qdisc *next;
};

struct sel_ctx {
    struct sel_path *paths;
    struct sel_ref *refs;
    struct sel_qdisc *qdiscs;
};

struct sel_prev {
    CLASS *class;
    const struct sel_prev *prev;
};


static void begin_selection_paths(struct sel_ctx *ctx)
{
    debugf("begin_selection_paths");
    ctx->paths = NULL;
    ctx->refs = NULL;
    ctx->qdiscs = NULL;
}


static void register_sel_class(struct sel_ctx *ctx,QDISC *qdisc,CLASS *class)
{
    struct sel_qdisc *q;
    struct sel_class *cl;

    for (q = ctx->qdiscs; q; q = q->next)
	if (q->qdisc == qdisc) break;
    if (!q) {
	q = alloc_t(struct sel_qdisc);
	q->qdisc = qdisc;
	q->classes = NULL;
	q->num_classes = 0;
	q->next = ctx->qdiscs;
	ctx->qdiscs = q;
    }
    for (cl = q->classes; cl; cl = cl->next)
	if (cl->class == class) return;
    cl = alloc_t(struct sel_class);
    cl->class = class;
    if (qdisc->dsc == &gred_dsc) cl->number = class->number;
    else cl->number = ++q->num_classes;
    cl->next = q->classes;
    q->classes = cl;
}


static void register_sel_path(struct sel_ctx *ctx,const struct sel_prev *last,
  LOCATION loc)
{
    struct sel_path_elem *path = NULL;
    struct sel_path *walk,*new;
    const struct sel_prev *prev;
    int n;

    for (prev = last; prev; prev = prev->prev) {
	QDISC *qdisc = prev->class->parent.qdisc;
	struct sel_path_elem *new;

	if (qdisc->filters)
	    lerror(qdisc->location,
	      "qdisc must not have filters when using class selection path");
	register_sel_class(ctx,qdisc,prev->class);
	new = alloc_t(struct sel_path_elem);
	new->class = prev->class;
	new->next = path;
	path = new;
    }
    n = 0;
    for (walk = ctx->paths; walk; walk = walk->next) {
	const struct sel_path_elem *a,*b;

	a = walk->path;
	b = path;
	while (a && b && a->class == b->class) {
	    a = a->next;
	    b = b->next;
	}
	if (!a && !b) {
	    n++;
	    walk->equals++;
	}
    }
//    if (walk) lerror(loc,"duplicate class selection path");
    new = alloc_t(struct sel_path);
    new->path = path;
    new->equals = n+1;
    new->index = n;
    new->next = ctx->paths;
    ctx->paths = new;
}


static void find_last_class(LOCATION loc,struct sel_ctx *ctx,
  const DATA_LIST *list,CLASS *classes,const CLASS **last_class,int left,
  const struct sel_prev *prev)
{
    CLASS *class;

    for (class = classes; class; class = class->next) {
	struct sel_prev this;
	const DATA_LIST *scan;

	this.class = class;
	this.prev = prev;
	for (scan = list; scan; scan = scan->next)
	    if (scan->ref->u.class == class) break;
	if (!scan) {
	    if (class->qdisc)
		find_last_class(loc,ctx,list,class->qdisc->classes,last_class,
		  left,&this);
	}
	else {
	    if (left == 1) {
		if (*last_class && *last_class != class)
		    lerror(loc,"class selection path is not unique");
		*last_class = class;
		register_sel_path(ctx,&this,loc);
		continue;
	    }
	    if (!class->qdisc)
		lerror(loc,"unfinished class selection path reaches leaf");
	    find_last_class(loc,ctx,list,class->qdisc->classes,last_class,
	      left-1,&this);
	}
	if (class->child)
	    find_last_class(loc,ctx,list,class->child,last_class,left,prev);
    }
}


static void register_selection_path(struct sel_ctx *ctx,
  uint32_t *number,const DATA_LIST *list,QDISC *inner_qdisc,LOCATION loc)
{
    const DATA_LIST *walk;
    const CLASS *last_class = NULL;
    struct sel_ref *ref;
    int classes = 0;

    for (walk = list; walk; walk = walk->next) {
	const DATA_LIST *scan;
	CLASS *class;

	if (walk->ref->type != dt_class)
	    lerror(loc,"invalid type in class selection path");
	class = walk->ref->u.class;
	for (scan = list; scan != walk; scan = scan->next)
	    if (scan->ref->u.class == class) {
		lwarn(loc,"duplicate class in class selection path");
		break;
	    }
	if (scan == walk) classes++;
    }
    find_last_class(loc,ctx,list,inner_qdisc->classes,&last_class,classes,NULL);
    if (!last_class) lerror(loc,"cannot resolve class selection path");
    ref= alloc_t(struct sel_ref);
    ref->number = number;
    ref->list = list;
    ref->next = ctx->refs;
    ctx->refs = ref;
}


static int selection_bits(struct sel_ctx *ctx,const QDISC *qdisc);


static int max_bits(struct sel_ctx *ctx,const CLASS *classes)
{
    const CLASS *class;
    int max = 0;

    for (class = classes; class; class = class->next) {
	int bits;

	if (class->qdisc) {
	    bits = selection_bits(ctx,class->qdisc);
	    if (bits > max) max = bits;
	}
	if (class->child) {
	    bits = max_bits(ctx,class->child);
	    if (bits > max) max = bits;
	}
    }
    return max;
}


/*
 * Bit assignment rules: we need enough bits for all classes that are selected
 * in a qdisc. Furthermore, one value (0) is reserved for not selecting any
 * class, whether this case may occur or not.
 */


static int selection_bits(struct sel_ctx *ctx,const QDISC *qdisc)
{
    struct sel_qdisc *q;

    for (q = ctx->qdiscs; q && q->qdisc != qdisc; q = q->next);
    if (!q) return 0; /* qdisc not used anywhere */
    if (qdisc->dsc == &gred_dsc) q->bits = 4;
    else for (q->bits = 0; 1 << q->bits <= q->num_classes; q->bits++);
    q->shift = max_bits(ctx,qdisc->classes);
    return q->shift+q->bits;
}


static FILTER *sel_add_tcindex(QDISC *qdisc,int shift,int mask_bits)
{
    FILTER **f;
    int n = 0;

    for (f = &qdisc->filters; *f; f = &(*f)->next) n++;
    *f = alloc_t(FILTER);
    (*f)->parent.device = qdisc->parent.device;
    (*f)->parent.qdisc = qdisc;
    (*f)->parent.class = NULL;
    (*f)->parent.filter = NULL;
    (*f)->parent.tunnel = NULL;
    (*f)->number = n+1;
    debugf("sel_add_tcindex: qdisc %u shift %d mask_bits %d (filter %d)",
      (unsigned) qdisc->number,shift,mask_bits,n+1);
    (*f)->location = qdisc->location;
    (*f)->dsc = &tcindex_dsc;
    (*f)->elements = NULL;
    (*f)->params = param_make(&prm_shift,data_unum(shift));
    (*f)->params->next =
      param_make(&prm_mask,data_unum(((1 << mask_bits)-1) << shift));
    (*f)->next = NULL;
    return *f;
}


static void sel_add_element(FILTER *f,CLASS *class,uint32_t n)
{
    ELEMENT *e;
    PARENT parent;

    debugf("sel_add_element: filter %u class %u:%u element %u",
      (unsigned) f->number,(unsigned) class->parent.qdisc->number,
      (unsigned) class->number,(unsigned) n);
    e = alloc_t(ELEMENT);
    parent = f->parent;
    parent.class = class;
    parent.filter = f;
    add_element(e,parent);
    e->number = n;
    e->location = f->location;
    e->police = NULL;
    e->next = NULL;
    e->params = NULL;
}


static void sel_add_hier_filters(struct sel_ctx *ctx)
{
    struct sel_qdisc *q;

    for (q = ctx->qdiscs; q; q = q->next) {
	FILTER *f;
	struct sel_class *cl;

	if (q->qdisc->dsc == &gred_dsc) continue;
	f = sel_add_tcindex(q->qdisc,q->shift,q->bits);
	for (cl = q->classes; cl; cl = cl->next)
	    sel_add_element(f,cl->class,cl->number);
    }
}


static void sel_set_class_numbers(struct sel_ctx *ctx,int top_shift)
{
    struct sel_ref *ref = ctx->refs;
    struct sel_path *path = ctx->paths;

    while (ref) {
	struct sel_path_elem *elem;
	uint32_t number;

	number = path->index << top_shift;
	for (elem = path->path; elem; elem = elem->next) {
	    const struct sel_qdisc *q;
	    const struct sel_class *cl;

	    for (q = ctx->qdiscs; q->qdisc != elem->class->parent.qdisc;
	      q = q->next);
	    for (cl = q->classes; cl->class != elem->class; cl = cl->next);
	    number |= cl->number << q->shift;
	}
	*ref->number = number;
	debugf("sel_set_class_numbers(%u)",(unsigned) number);
	ref = ref->next;
	path = path->next;
    }
}


static void set_number_refs(struct sel_ctx *ctx)
{
    struct sel_ref *ref;
    int n = 0;

    for (ref = ctx->refs; ref; ref = ref->next)
	*ref->number = ++n;
    if (n >= 1 << DSMARK_INDEX_BITS)
	errorf("number of class selection paths exceeds %u",
	  1 << DSMARK_INDEX_BITS);
}


static void sel_add_seq_filters(struct sel_ctx *ctx)
{
    const struct sel_qdisc *q;
    struct sel_path *path = ctx->paths;
    int n = 0;

    for (q = ctx->qdiscs; q; q = q->next)
	if (q->qdisc->dsc != &gred_dsc)
	    (void) sel_add_tcindex(q->qdisc,0,DSMARK_INDEX_BITS);
    for (path = ctx->paths; path; path = path->next) {
	struct sel_path_elem *elem;

	n++;
	for (elem = path->path; elem; elem = elem->next) {
	    const QDISC *qdisc = elem->class->parent.qdisc;
	    const struct sel_class *cl;

	    if (qdisc->dsc == &gred_dsc) continue;
	    for (q = ctx->qdiscs; q->qdisc != qdisc; q = q->next);
	    for (cl = q->classes; cl->class != elem->class; cl = cl->next);
	    sel_add_element(q->qdisc->filters,cl->class,n);
	}
    }
}


static void sel_destroy(struct sel_ctx *ctx)
{
    while (ctx->paths) {
	struct sel_path *next = ctx->paths->next;

	while (ctx->paths->path) {
	    struct sel_path_elem *next = ctx->paths->path->next;

	    free(ctx->paths->path);
	    ctx->paths->path = next;
	}
	free(ctx->paths);
	ctx->paths = next;
    }
    while (ctx->refs) {
	struct sel_ref *next = ctx->refs->next;

	free(ctx->refs);
	ctx->refs = next;
    }
    while (ctx->qdiscs) {
	struct sel_qdisc *next = ctx->qdiscs->next;

	while (ctx->qdiscs->classes) {
	    struct sel_class *next = ctx->qdiscs->classes->next;

	    free(ctx->qdiscs->classes);
	    ctx->qdiscs->classes = next;
	}
	free(ctx->qdiscs);
	ctx->qdiscs = next;
    }
}


static void end_selection_paths(struct sel_ctx *ctx,QDISC *inner_qdisc)
{
    const struct sel_path *path;
    int max_equals,top_bits,bits;

    debugf("end_selection_paths (%p)",ctx->paths);
    if (!ctx->paths) return;
    max_equals = 0;
    for (path = ctx->paths; path; path = path->next)
	if (path->equals > max_equals) max_equals = path->equals;
    for (top_bits = 0; (1 << top_bits) < max_equals; top_bits++);
    bits = selection_bits(ctx,inner_qdisc);
    if (top_bits+bits <= DSMARK_INDEX_BITS) {
	sel_add_hier_filters(ctx);
	sel_set_class_numbers(ctx,bits);
    }
    else {
	lwarnf(inner_qdisc->location,
	  "number of index bits required for path exceeds %d",
	  DSMARK_INDEX_BITS);
	set_number_refs(ctx);
	sel_add_seq_filters(ctx);
    }
    sel_destroy(ctx);
}


/* ----- Checking ---------------------------------------------------------- */


static int merge_class_zero(CLASS **classes)
{
    CLASS **default_zero = NULL,*explicit_zero = NULL;
    CLASS **class,*tmp;

    for (class = classes; *class; class = &(*class)->next) {
	if ((*class)->qdisc && (*class)->number == UNDEF_U32)
	    default_zero = class;
	if (!(*class)->number) explicit_zero = *class;
    }
    if (!default_zero || !explicit_zero) return 0;
    explicit_zero->qdisc = (*default_zero)->qdisc;
    explicit_zero->qdisc->parent.class = explicit_zero;
    tmp = *default_zero;
    *default_zero = tmp->next;
    free(tmp);
    return 1;
}


static void dsmark_check(QDISC *qdisc)
{
    DEFAULT_DECL;
    QDISC *inner_qdisc = NULL;
    CLASS *class;
    struct sel_ctx ctx;
    int merged,have_indices;
    uint32_t indices;

    param_get(qdisc->params,qdisc->location);
    DEFAULT_SET;
    have_indices = prm_indices.present;
    indices = prm_indices.v;
    if (have_indices) {
	int tmp;

	for (tmp = prm_indices.v; tmp != 1; tmp >>= 1)
	    if (tmp & 1)
		lerror(qdisc->location,"indices must be a power of two");
    }
    merged = merge_class_zero(&qdisc->classes);

    /*
     * Pass 1: check values, find inner qdisc, and check it (assign class
     * numbers, etc.)
     */
    for (class = qdisc->classes; class; class = class->next) {
	check_childless(class);
	param_get(class->params,class->location);

	/* The following test must be before DEFAULT_GET ! */
	if (class->qdisc && (class->number != UNDEF_U32 ||
	  prm_mask.present || prm_value.present || prm_class_sel_path.present)
	  && !merged)
	    lwarn(class->location,
	      "suggest unnumbered and parameter-less class for dsmark's inner "
	      "qdisc");
	DEFAULT_GET;
	if (prm_mask.present && prm_mask.v > 0xff)
	    lerrorf(class->location,"mask 0x%x above limit 0xff",prm_mask.v);
	if (!class->qdisc) continue;
	if (inner_qdisc)
	    lerror(class->location,"dsmark has only one inner qdisc");
	check_qdisc(class->qdisc);
	inner_qdisc = class->qdisc;
    }

    /*
     * Pass 2: collect class selection paths, and invoke class selection path
     * magic.
     */
    begin_selection_paths(&ctx);
    for (class = qdisc->classes; class; class = class->next) {
	param_get(class->params,class->location);
	if (prm_class_sel_path.present) {
	    if (class->number != UNDEF_U32)
		lerror(class->location,"can't use class selection path and "
		  "class number on same class");
	    if (!inner_qdisc)
		lerror(class->location,"without an inner qdisc, there is "
		  "scarcely anything to select");
	    register_selection_path(&ctx,&class->number,
	      prm_data(class->params,&prm_class_sel_path).u.list,inner_qdisc,
	      class->location);
	}
	else {
	    if (class->number == UNDEF_U32) {
		if (class->qdisc) class->number = 0;
		else lerror(class->location,
		      "dsmark does not auto-assign class numbers");
	    }
	}
    }
    end_selection_paths(&ctx,inner_qdisc);

    /*
     * Pass 3: check class number range
     */
    for (class = qdisc->classes; class; class = class->next) {
	param_get(class->params,class->location);
	DEFAULT_GET;
	if (((prm_mask.present && prm_mask.v != 0xff) ||
	  (prm_value.present && prm_value.v)) && have_indices &&
	  class->number >= indices)
	    lerrorf(class->location,
	      "dsmark class number (%lu) must be < indices (%lu) if using "
	      "parameters",(unsigned long) class->number,
	      (unsigned long) indices-1);
    }
    sort_classes(&qdisc->classes);
    check_filters(qdisc->filters);
}


/* ----- Default classification -------------------------------------------- */


/*
 * @@@ Supporting dsmark's behaviour in the absence of a default index is
 * kind of difficult, because there are 64k possible skb->tc_index values
 * that need to be mapped to "classes". So it's better to wait with this
 * until we can absorb dsmark into some more abstract data forwarding
 * model.
 *
 * Update: at least if set_tc_index is present, things don't look quite so
 * bad.
 */

static void dsmark_default_class(DATA *d,QDISC *qdisc)
{
    DATA dfl;
    int i;

    param_get(qdisc->params,qdisc->location);
    if (prm_default_index.present) {
	append_default(d,
	  data_decision(dr_class,require_class(qdisc,prm_default_index.v)));
	return;
    }
    if (!prm_set_tc_index.present || !prm_set_tc_index.v)
	lerror(qdisc->location,"dsmark_default_class: need "
	  "\"default_index\" or \"set_tc_index\"");
    dfl = data_decision(dr_class,require_class(qdisc,255));
    for (i = 254; i >= 0; i--)
	dfl = op_binary(&op_logical_or,
	  op_binary(&op_logical_and,
	    op_binary(&op_eq,
	      op_ternary(&op_access,data_none(),data_unum(1),data_unum(8)),
	      data_unum(i)),
	    data_decision(dr_class,require_class(qdisc,i))),
          dfl);
    append_default(d,dfl);
}


/* ----- Helper function for setting "indices" ----------------------------- */


static uint32_t get_indices(const QDISC *qdisc)
{
    const CLASS *class;
    uint32_t max_class = 0,indices;

    if (prm_indices.present) return prm_indices.v;
    for (class = qdisc->classes; class; class = class->next)
	if (class->number > max_class) max_class = class->number;
    for (indices = 1; indices <= max_class; indices *= 2);
    return indices;
}


/* ----- Dump tc ----------------------------------------------------------- */


static void dsmark_dump_tc(QDISC *qdisc)
{
    DEFAULT_DECL;
    CLASS *class;

    tc_pragma(qdisc->params);
    param_get(qdisc->params,qdisc->location);
    DEFAULT_SET;
    tc_qdisc_add(qdisc);
    tc_more(" indices %d",(int) get_indices(qdisc));
    if (prm_default_index.present)
	tc_add_unum("default_index",prm_default_index.v);
    if (prm_set_tc_index.present) tc_more(" set_tc_index");
    tc_nl();
    for (class = qdisc->classes; class; class = class->next) {
	tc_pragma(class->params);
	param_get(class->params,class->location);
	DEFAULT_GET;
	if (prm_mask.present || prm_value.present) {
	    tc_class_change(class);
	    if (prm_mask.present) tc_add_hex("mask",prm_mask.v);
	    if (prm_value.present) tc_add_hex("value",prm_value.v);
	    tc_nl();
	}
	dump_qdisc(class->qdisc);
    }
    dump_filters(qdisc->filters);
}


/* ----- Dump external ----------------------------------------------------- */


static void print_indices(FILE *file,const PARAM_DSC *param,const QDISC *qdisc)
{
    if (param == &prm_indices)
	fprintf(file," indices %d",(int) get_indices(qdisc));
}


static void dsmark_dump_ext(FILE *file,QDISC *qdisc)
{
    const PARAM_DSC *indices[] = {
	&prm_indices,
	NULL
    };
    const PARAM_DSC *class_sel_path[] = {
	&prm_class_sel_path,
	NULL
    };
    const CLASS *class;

    qdisc_param_load(qdisc);
    ext_dump_qdisc(file,qdisc,indices,print_indices);
    for (class = qdisc->classes; class; class = class->next) {
        param_push();
        class_param_load(class);
        ext_dump_class(file,class,class_sel_path,NULL);
        param_pop();
    }

}


/* ----- Descriptors ------------------------------------------------------- */


static const PARAM_DSC *dsmark_qdisc_opt[] = {
    &prm_default_index,
    &prm_indices,
    &prm_pragma,	/* list */
    &prm_set_tc_index,
    DEFAULT_LIST	/* NB: no trailing comma */
    NULL
};

static const PARAM_DSC *dsmark_class_opt[] = {
    &prm_class_sel_path,
    &prm_pragma,	/* list */
    &prm_value,
    DEFAULT_LIST	/* NB: no trailing comma */
    NULL
};

static PARAM_DEF dsmark_qdisc = {
    .required = NULL,
    .optional = dsmark_qdisc_opt,
};

static PARAM_DEF dsmark_class = {
    .required = NULL,
    .optional = dsmark_class_opt,
};

QDISC_DSC dsmark_dsc = {
    .name = "dsmark",
    .flags = QDISC_HAS_CLASSES | QDISC_HAS_FILTERS | QDISC_HAS_DEFAULT |
      QDISC_CHILD_QDISCS | QDISC_SHARED_QDISC,
    .qdisc_param = &dsmark_qdisc,
    .class_param = &dsmark_class,
    .check = &dsmark_check,
    .default_class = &dsmark_default_class,
    .dump_tc = &dsmark_dump_tc,
    .dump_ext = &dsmark_dump_ext,
};
